/*
  VisualHash - A Visual Hash Calculator
  Copyright (C) 2005 Dominik Reichl <oss@dominik-reichl.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

using System;

namespace VisualHash.ReHashLib
{
	public class MD2Hash : ReHashModule
	{
		private byte[] m_X = new byte[48];
		private byte[] m_C = new byte[16];
		private byte[] m_buf = new byte[16];
		private uint m_count = 0;

		// MD2 S-Box, based on PI
		private static byte[] MD2_S = {
			41, 46, 67, 201, 162, 216, 124, 1, 61, 54, 84, 161, 236, 240, 6,
			19, 98, 167, 5, 243, 192, 199, 115, 140, 152, 147, 43, 217, 188,
			76, 130, 202, 30, 155, 87, 60, 253, 212, 224, 22, 103, 66, 111, 24,
			138, 23, 229, 18, 190, 78, 196, 214, 218, 158, 222, 73, 160, 251,
			245, 142, 187, 47, 238, 122, 169, 104, 121, 145, 21, 178, 7, 63,
			148, 194, 16, 137, 11, 34, 95, 33, 128, 127, 93, 154, 90, 144, 50,
			39, 53, 62, 204, 231, 191, 247, 151, 3, 255, 25, 48, 179, 72, 165,
			181, 209, 215, 94, 146, 42, 172, 86, 170, 198, 79, 184, 56, 210,
			150, 164, 125, 182, 118, 252, 107, 226, 156, 116, 4, 241, 69, 157,
			112, 89, 100, 113, 135, 32, 134, 91, 207, 101, 230, 45, 168, 2, 27,
			96, 37, 173, 174, 176, 185, 246, 28, 70, 97, 105, 52, 64, 126, 15,
			85, 71, 163, 35, 221, 81, 175, 58, 195, 92, 249, 206, 186, 197,
			234, 38, 44, 83, 13, 110, 133, 40, 132, 9, 211, 223, 205, 244, 65,
			129, 77, 82, 106, 220, 55, 200, 108, 193, 171, 250, 36, 225, 123,
			8, 12, 189, 177, 74, 120, 136, 149, 139, 227, 99, 232, 109, 233,
			203, 213, 254, 59, 0, 29, 57, 242, 239, 183, 14, 102, 88, 208, 228,
			166, 119, 114, 248, 235, 117, 75, 10, 49, 68, 80, 180, 143, 237,
			31, 26, 219, 153, 141, 51, 159, 17, 131, 20
		};

		public MD2Hash()
		{
		}

		public override bool InitModule()
		{
			strName = "MD2";
			strShortName = "MD2";
			strRef = "RFC 1319";
			uHashSize = 128;
			uInternalBufferSize = 128;

			pbFinalHash = new byte[16];
			return true;
		}

		public override void ReleaseModule()
		{
			SecureZeroArray(pbFinalHash);
			pbFinalHash = null;
		}

		public override void InitNewHash()
		{
			uint i;
			for(i = 0; i < 48; i++) m_X[i] = 0;
			for(i = 0; i < 16; i++) m_C[i] = 0;
			for(i = 0; i < 16; i++) m_buf[i] = 0;
			
			m_count = 0;
		}

		public override void UpdateHash(byte[] pbData, ulong uLength, bool bIsLastBlock)
		{
			uint L, uDataInx = 0;
			byte t, i, j;

			while(uLength != 0)
			{
				L = ((16 - m_count) < uLength) ? (16 - m_count) : (uint)uLength;
				CopyMemoryEx(m_buf, m_count, pbData, uDataInx, L);
				m_count += L;
				uDataInx += L;
				uLength -= L;

				if(m_count == 16)
				{
					m_count = 0;
					CopyMemoryEx(m_X, 16, m_buf, 0, 16);
					t = m_C[15];

					for(i = 0; i < 16; i++)
					{
						m_X[32 + i] = (byte)(m_X[16 + i] ^ m_X[i]);
						t = m_C[i] ^= MD2_S[m_buf[i] ^ t];
					}

					t = 0;
					for(i = 0; i < 18; i++)
					{
						for(j = 0; j < 48; j++)
							t = m_X[j] ^= MD2_S[t];

						t += i;
					}
				}
			}
		}

		public override void FinalizeHash()
		{
			byte[] aPadding = new byte[16];
			byte padLen = (byte)(16 - m_count);

			for(uint i = 0; i < padLen; i++)
				aPadding[i] = padLen;

			UpdateHash(aPadding, padLen, false);
			UpdateHash(m_C, 16, true);

			CopyMemoryEx(pbFinalHash, 0, m_X, 0, 16);
		}

		public override uint Test()
		{
			byte[] pb;
			byte[] r1 = { 0x9F, 0xF6, 0x31, 0x9C, 0x34, 0x4E, 0x27, 0x7D,
				0xCC, 0xF5, 0xFD, 0x89, 0xBC, 0x7A, 0xFC, 0x91 };
			byte[] r2 = { 0x83, 0x50, 0xe5, 0xa3, 0xe2, 0x4c, 0x15, 0x3d,
				0xf2, 0x27, 0x5c, 0x9f, 0x80, 0x69, 0x27, 0x73 };
			byte[] r3 = { 0x32, 0xec, 0x01, 0xec, 0x4a, 0x6d, 0xac, 0x72,
				0xc0, 0xab, 0x96, 0xfb, 0x34, 0xc0, 0xb5, 0xd1 };
			byte[] r4 = { 0xd5, 0x97, 0x6f, 0x79, 0xd8, 0x3d, 0x3a, 0x0d,
				0xc9, 0x80, 0x6c, 0x3c, 0x66, 0xf3, 0xef, 0xd8 };

			pb = StringToAsciiEx("abcabcabcabcabcabcabcabcabcabcabcabcabcabcabc");
			if(!HashAndCompare(pb, r1)) return 1;

			InitNewHash();
			UpdateHash(pbFinalHash, 0, true);
			FinalizeHash();
			if(!CompareByteArraysEx(pbFinalHash, r2)) return 2;
			
			pb = StringToAsciiEx("a");
			if(!HashAndCompare(pb, r3)) return 3;

			pb = StringToAsciiEx("12345678901234567890123456789012345678901234567890123456789012345678901234567890");
			if(!HashAndCompare(pb, r4)) return 4;
			
			return 0;
		}
	}
}
