/*
  VisualHash - A Visual Hash Calculator
  Copyright (C) 2005 Dominik Reichl <oss@dominik-reichl.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

using System;

namespace VisualHash.ReHashLib
{
	public class MD4Hash : ReHashModule
	{
		uint[] m_state = new uint[4];
		uint[] m_count = new uint[2];
		byte[] m_buffer = new byte[64];

		private const int MD4_S11 = 3;
		private const int MD4_S12 = 7;
		private const int MD4_S13 = 11;
		private const int MD4_S14 = 19;
		private const int MD4_S21 = 3;
		private const int MD4_S22 = 5;
		private const int MD4_S23 = 9;
		private const int MD4_S24 = 13;
		private const int MD4_S31 = 3;
		private const int MD4_S32 = 9;
		private const int MD4_S33 = 11;
		private const int MD4_S34 = 15;

		private byte[] MD4_PADDING = {
			0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
		};

		private static void MD4_FF(ref uint a, uint b, uint c, uint d, uint x, int s)
		{
			a += ((b & c) | ((~b) & d)) + x;
			a = (a << s) | (a >> (32 - s)); // a = ROL32((a), (s));
		}

		private static void MD4_GG(ref uint a, uint b, uint c, uint d, uint x, int s)
		{
			a += ((b & c) | (b & d) | (c & d)) + x + 0x5a827999u;
			a = (a << s) | (a >> (32 - s)); // a = ROL32((a), (s));
		}

		private static void MD4_HH(ref uint a, uint b, uint c, uint d, uint x, int s)
		{
			a += (b ^ c ^ d) + x + 0x6ed9eba1u;
			a = (a << s) | (a >> (32 - s)); // a = ROL32((a), (s));
		}

		private static void Transform(uint[] pState, byte[] pBlock, uint uBlockOffset)
		{
			uint a = pState[0], b = pState[1], c = pState[2], d = pState[3];
			uint[] x = new uint[16];

			Decode(x, pBlock, uBlockOffset, 64);

			// Round 1
			MD4_FF(ref a, b, c, d, x[ 0], MD4_S11);
			MD4_FF(ref d, a, b, c, x[ 1], MD4_S12);
			MD4_FF(ref c, d, a, b, x[ 2], MD4_S13);
			MD4_FF(ref b, c, d, a, x[ 3], MD4_S14);
			MD4_FF(ref a, b, c, d, x[ 4], MD4_S11);
			MD4_FF(ref d, a, b, c, x[ 5], MD4_S12);
			MD4_FF(ref c, d, a, b, x[ 6], MD4_S13);
			MD4_FF(ref b, c, d, a, x[ 7], MD4_S14);
			MD4_FF(ref a, b, c, d, x[ 8], MD4_S11);
			MD4_FF(ref d, a, b, c, x[ 9], MD4_S12);
			MD4_FF(ref c, d, a, b, x[10], MD4_S13);
			MD4_FF(ref b, c, d, a, x[11], MD4_S14);
			MD4_FF(ref a, b, c, d, x[12], MD4_S11);
			MD4_FF(ref d, a, b, c, x[13], MD4_S12);
			MD4_FF(ref c, d, a, b, x[14], MD4_S13);
			MD4_FF(ref b, c, d, a, x[15], MD4_S14);
		
			// Round 2
			MD4_GG(ref a, b, c, d, x[ 0], MD4_S21);
			MD4_GG(ref d, a, b, c, x[ 4], MD4_S22);
			MD4_GG(ref c, d, a, b, x[ 8], MD4_S23);
			MD4_GG(ref b, c, d, a, x[12], MD4_S24);
			MD4_GG(ref a, b, c, d, x[ 1], MD4_S21);
			MD4_GG(ref d, a, b, c, x[ 5], MD4_S22);
			MD4_GG(ref c, d, a, b, x[ 9], MD4_S23);
			MD4_GG(ref b, c, d, a, x[13], MD4_S24);
			MD4_GG(ref a, b, c, d, x[ 2], MD4_S21);
			MD4_GG(ref d, a, b, c, x[ 6], MD4_S22);
			MD4_GG(ref c, d, a, b, x[10], MD4_S23);
			MD4_GG(ref b, c, d, a, x[14], MD4_S24);
			MD4_GG(ref a, b, c, d, x[ 3], MD4_S21);
			MD4_GG(ref d, a, b, c, x[ 7], MD4_S22);
			MD4_GG(ref c, d, a, b, x[11], MD4_S23);
			MD4_GG(ref b, c, d, a, x[15], MD4_S24);
		
			// Round 3
			MD4_HH(ref a, b, c, d, x[ 0], MD4_S31);
			MD4_HH(ref d, a, b, c, x[ 8], MD4_S32);
			MD4_HH(ref c, d, a, b, x[ 4], MD4_S33);
			MD4_HH(ref b, c, d, a, x[12], MD4_S34);
			MD4_HH(ref a, b, c, d, x[ 2], MD4_S31);
			MD4_HH(ref d, a, b, c, x[10], MD4_S32);
			MD4_HH(ref c, d, a, b, x[ 6], MD4_S33);
			MD4_HH(ref b, c, d, a, x[14], MD4_S34);
			MD4_HH(ref a, b, c, d, x[ 1], MD4_S31);
			MD4_HH(ref d, a, b, c, x[ 9], MD4_S32);
			MD4_HH(ref c, d, a, b, x[ 5], MD4_S33);
			MD4_HH(ref b, c, d, a, x[13], MD4_S34);
			MD4_HH(ref a, b, c, d, x[ 3], MD4_S31);
			MD4_HH(ref d, a, b, c, x[11], MD4_S32);
			MD4_HH(ref c, d, a, b, x[ 7], MD4_S33);
			MD4_HH(ref b, c, d, a, x[15], MD4_S34);

			pState[0] += a; pState[1] += b;
			pState[2] += c; pState[3] += d;
		}

		private static void Encode(byte[] pOutput, uint[] pInput, uint uLen)
		{
			uint i, j, l;

			for(i = 0, j = 0; j < uLen; i++, j += 4)
			{
				l = pInput[i];
				pOutput[j] = (byte)(l & 0xff);
				pOutput[j+1] = (byte)((l >> 8) & 0xff);
				pOutput[j+2] = (byte)((l >> 16) & 0xff);
				pOutput[j+3] = (byte)(l >> 24);
			}
		}

		private static void Decode(uint[] pOutput, byte[] pInput, uint uInputOffset, uint uLen)
		{
			uint i, j;

			for(i = 0, j = uInputOffset; j < (uLen + uInputOffset); i++, j += 4)
				pOutput[i] = ((uint)pInput[j]) | (((uint)pInput[j+1]) << 8) |
					(((uint)pInput[j+2]) << 16) | (((uint)pInput[j+3]) << 24);
		}

		public MD4Hash()
		{
		}

		public override bool InitModule()
		{
			strName = "MD4";
			strShortName = "MD4";
			strRef = "RFC 1320";
			uHashSize = 128;
			uInternalBufferSize = 512;

			pbFinalHash = new byte[16];
			return true;
		}

		public override void ReleaseModule()
		{
			SecureZeroArray(pbFinalHash);
			pbFinalHash = null;
		}

		public override void InitNewHash()
		{
			m_count[0] = m_count[1] = 0u;

			// Load magic initialization constants
			m_state[0] = 0x67452301u;
			m_state[1] = 0xefcdab89u;
			m_state[2] = 0x98badcfeu;
			m_state[3] = 0x10325476u;
		}

		public override void UpdateHash(byte[] pbData, ulong uLength, bool bIsLastBlock)
		{
			uint i = 0, ix, partLen;
			uint uLen = (uint)uLength;

			// Compute number of bytes mod 64
			ix = (m_count[0] >> 3) & 0x3f;

			// Update number of bits
			m_count[0] += uLen << 3;
			if(m_count[0] < (uLen << 3)) m_count[1]++;
			m_count[1] += uLen >> 29;

			partLen = 64 - ix;

			// Transform as many times as possible
			if(uLen >= partLen)
			{
				CopyMemoryEx(m_buffer, ix, pbData, 0, partLen);
				Transform(m_state, m_buffer, 0);

				for(i = partLen; (i + 63) < uLen; i += 64)
					Transform(m_state, pbData, i);

				ix = 0;
			}
			else i = 0;
			
			// Buffer remaining input
			CopyMemoryEx(m_buffer, ix, pbData, i, uLen - i);
		}

		public override void FinalizeHash()
		{
			byte[] pBits = new byte[8];
			uint ix, padLen;

			// Save number of bits
			Encode(pBits, m_count, 8);
			
			// Pad out to 56 mod 64.
			ix = (m_count[0] >> 3) & 0x0000003fu;
			padLen = (ix < 56) ? (56 - ix) : (120 - ix);
			UpdateHash(MD4_PADDING, padLen, false);

			// Append length (before padding)
			UpdateHash(pBits, 8, true);

			// Store state in digest
			Encode(pbFinalHash, m_state, 16);
		}

		public override uint Test()
		{
			byte[] pb;
			byte[] r1 = { 0xEA, 0xEA, 0x40, 0xDA, 0x1C, 0x96, 0xCF, 0x58,
				0xD6, 0xE9, 0x8B, 0xCA, 0xFD, 0x0C, 0xAA, 0x8B };
			byte[] r2 = { 0xe3, 0x3b, 0x4d, 0xdc, 0x9c, 0x38, 0xf2, 0x19,
				0x9c, 0x3e, 0x7b, 0x16, 0x4f, 0xcc, 0x05, 0x36 };
			byte[] r3 = { 0x31, 0xd6, 0xcf, 0xe0, 0xd1, 0x6a, 0xe9, 0x31,
				0xb7, 0x3c, 0x59, 0xd7, 0xe0, 0xc0, 0x89, 0xc0 };
			byte[] r4 = { 0xbd, 0xe5, 0x2c, 0xb3, 0x1d, 0xe3, 0x3e, 0x46,
				0x24, 0x5e, 0x05, 0xfb, 0xdb, 0xd6, 0xfb, 0x24 };

			pb = StringToAsciiEx("abcabcabcabcabcabcabcabcabcabcabcabcabcabcabc");
			if(!HashAndCompare(pb, r1)) return 1;
			
			pb = StringToAsciiEx("12345678901234567890123456789012345678901234567890123456789012345678901234567890");
			if(!HashAndCompare(pb, r2)) return 2;
			
			InitNewHash();
			UpdateHash(pbFinalHash, 0, true);
			FinalizeHash();
			if(!CompareByteArraysEx(pbFinalHash, r3)) return 3;
			
			pb = StringToAsciiEx("a");
			if(!HashAndCompare(pb, r4)) return 4;
			
			return 0;
		}
	}
}
