/*
  VisualHash - A Visual Hash Calculator
  Copyright (C) 2005 Dominik Reichl <oss@dominik-reichl.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

using System;

namespace VisualHash.ReHashLib
{
	public class XUM32Hash : ReHashModule
	{
		private uint m_xum32 = 0;
		private CRC32Hash m_crc32 = new CRC32Hash();
		private ELF32Hash m_elf32 = new ELF32Hash();

		public XUM32Hash()
		{
		}
		
		public override bool InitModule()
		{
			strName = "XUM-32";
			strShortName = "XUM-32";
			strRef = "";
			uHashSize = 32;
			uInternalBufferSize = 32;
			
			m_crc32.InitModule();
			m_elf32.InitModule();

			pbFinalHash = new byte[4];
			return true;
		}

		public override void ReleaseModule()
		{
			m_crc32.ReleaseModule();
			m_elf32.ReleaseModule();

			SecureZeroArray(pbFinalHash);
			pbFinalHash = null;
		}

		public override void InitNewHash()
		{
			m_crc32.InitNewHash();
			m_elf32.InitNewHash();

			m_xum32 = 0;
		}

		public override void UpdateHash(byte[] pbData, ulong uLength, bool bIsLastBlock)
		{
			m_crc32.UpdateHash(pbData, uLength, bIsLastBlock);
			m_elf32.UpdateHash(pbData, uLength, bIsLastBlock);

			m_xum32 += (uint)uLength;
		}

		public override void FinalizeHash()
		{
			uint uCRC = m_crc32.GetRawCRC32Value();
			uint uELF = m_elf32.GetRawELFValue();
			
			// Now allow inherited classes to clear their internal buffers
			m_crc32.FinalizeHash();
			m_elf32.FinalizeHash();

			uCRC ^= 0xFFFFFFFF;
			uELF %= 997;

			m_xum32 = (m_xum32 << 16) + (m_xum32 >> 16);

			m_xum32 ^= uCRC;
			m_xum32 ^= uELF;

			Store32H(pbFinalHash, m_xum32);
		}

		public override uint Test()
		{
			byte[] pb;
			byte[] r1 = { 0x2B, 0x75, 0xB9, 0x4C };

			pb = StringToAsciiEx("abcabcabcabcabcabcabcabcabcabcabcabcabcabcabc");
			if(!HashAndCompare(pb, r1)) return 1;

			return 0;
		}
	}
}
